/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 * 
 * Copyright (©) 2005,2006,2007,2008 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2002,2007 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992,2007 Frank Tore Johansen
 * 
 * Deliantra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

/**
 * \file
 * Client/server logic.
 *
 * \date 2003-12-02
 *
 * This contains item logic for client/server.  It doesn't contain
 * the actual commands that send the data, but does contain
 * the logic for what items should be sent.
 */

#include <global.h>
#include <object.h>
#include <sproto.h>

/** This is the maximum number of bytes we expect any one item to take up */
#define MAXITEMLEN  300

/*******************************************************************************
 *
 * Functions related to sending object data to the client.
 *
 ******************************************************************************/

/**
 *  This is a similar to query_name, but returns flags
 *  to be sent to client. 
 */
unsigned int
query_flags (object *op)
{
  unsigned int flags = 0;

  if (QUERY_FLAG (op, FLAG_APPLIED))
    switch (op->type)
      {
        case BOW:
        case WAND:
        case ROD:
        case HORN:
          flags = a_readied;
          break;
        case WEAPON:
          flags = a_wielded;
          break;
        case SKILL:
        case ARMOUR:
        case HELMET:
        case SHIELD:
        case RING:
        case BOOTS:
        case GLOVES:
        case AMULET:
        case GIRDLE:
        case BRACERS:
        case CLOAK:
          flags = a_worn;
          break;
        case CONTAINER:
          flags = a_active;
          break;
        default:
          flags = a_applied;
          break;
      }

  if (op->type == CONTAINER && ((op->env && op->env->container == op) || (!op->env && QUERY_FLAG (op, FLAG_APPLIED))))
    flags |= F_OPEN;

  if (QUERY_FLAG (op, FLAG_KNOWN_CURSED))
    {
      if (QUERY_FLAG (op, FLAG_DAMNED))
        flags |= F_DAMNED;
      else if (QUERY_FLAG (op, FLAG_CURSED))
        flags |= F_CURSED;
    }

  if (QUERY_FLAG (op, FLAG_KNOWN_MAGICAL) && !QUERY_FLAG (op, FLAG_IDENTIFIED))
    flags |= F_MAGIC;
  if (QUERY_FLAG (op, FLAG_UNPAID))
    flags |= F_UNPAID;
  if (QUERY_FLAG (op, FLAG_INV_LOCKED))
    flags |= F_LOCKED;

  return flags;
}

/* Used in the send_look to put object head into packet
 * sl for socket ns.  Need socket to know if we need to send
 * animation of face to the client.
 */
static void
add_object_to_socklist (client &ns, packet &sl, object *head)
{
  char item_n[MAX_BUF];
  const char *item_p;

  int flags = query_flags (head);
  if (QUERY_FLAG (head, FLAG_NO_PICK))
    flags |= F_NOPICK;

  ns.send_face (head->face, -50);
  ns.flush_fx ();

  if (QUERY_FLAG (head, FLAG_ANIMATE) && !ns.anims_sent[head->animation_id])
    ns.send_animation (head->animation_id);

  sl << uint32 (head->count)
     << uint32 (flags)
     << uint32 (QUERY_FLAG (head, FLAG_NO_PICK) ? -1 : head->client_weight ())
     << uint32 (head->face);

  int len;

  if (!head->custom_name)
    {
      strncpy (item_n, query_base_name (head, 0), 127);
      item_n[127] = 0;
      len = strlen (item_n);
      item_p = query_base_name (head, 1);
    }
  else
    {
      strncpy (item_n, head->custom_name, 127);
      item_n[127] = 0;
      len = strlen (item_n);
      item_p = head->custom_name;
    }

  strncpy (item_n + len + 1, item_p, 127);
  item_n[254] = 0;
  len += strlen (item_n + 1 + len) + 1;

  sl << data8 (item_n, len)
     << uint16 (head->animation_id);

  int anim_speed = !head->flag [FLAG_ANIMATE] ? 0
                 : head->anim_speed           ? clamp (head->anim_speed, 1, 255)
                 :                              1. / clamp (fabs (head->speed), 1./255., 1./1.);

  sl << uint8  (anim_speed)
     << uint32 (head->nrof);

  if (ns.itemcmd == 2)
    sl << uint16 (head->client_type);

  SET_FLAG (head, FLAG_CLIENT_SENT);
}

static faceidx
need_face_now (player *pl, const char *name)
{
  faceidx face = face_find (name, empty_face);

  pl->ns->send_face (face, -50);
  pl->ns->flush_fx ();

  return face;
}

#define FINGER_UP   "finger_up.x11"
#define FINGER_DOWN "finger_down.x11"

/**
 * Send the look window. Don't need to do animations here 
 * This sends all the faces to the client, not just updates. This is
 * because object ordering would otherwise be inconsistent.
 */
void
esrv_draw_look (player *pl)
{
  object *ob = pl->ob;

  if (!pl->ns->update_look)
    {
      LOG (llevDebug, "esrv_draw_look called when update_look was not set (player %s)\n", &ob->name);
      return;
    }

  pl->ns->update_look = 0;

  if (QUERY_FLAG (ob, FLAG_REMOVED)
      || !ob->map
      || ob->map->in_memory != MAP_ACTIVE
      || out_of_map (ob->map, ob->x, ob->y))
    return;

  pl->ns->send_packet ("delinv 0");

  packet sl;
  sl.printf ("item%d ", pl->ns->itemcmd);

  sl << uint32 (0);

  int start_pos = pl->ns->look_position;
  bool dirty = false;

  mapspace &ms = ob->ms ();

  // manage a ring buffer of the "last FLOORBOX_PAGESIZE" items and
  // start from the top
  object *items [FLOORBOX_PAGESIZE];
  int item_idx = 0, item_cnt = 0;
  int pos = 0;

  // find our items by walking down
  object *item = ms.top;

  for (; item && item_cnt < FLOORBOX_PAGESIZE; item = item->below)
    {
      if (!item->client_visible ())
        continue;

      if (++pos < start_pos)
        continue;

      // record item
      items [item_idx] = item; item_idx = item_idx < FLOORBOX_PAGESIZE ? item_idx + 1 : 0;
      item_cnt++;

      // stop at first floor
      if (item->flag [FLAG_IS_FLOOR])
        {
          // we are finished, don't append "next group of items"
          // by setting item to ms.bot it becomes zero which is checked after the while loop
          item = ms.bot;
        }
    }

  // see if there are more if we cared - we ignore invisible objects
  if (item)
    {
      /* What we basically do is make a 'fake' object - when the user applies it,
       * we notice the special tag the object has, and act accordingly.
       */
      sl << uint32 (0x80000000 | (start_pos + FLOORBOX_PAGESIZE))
         << uint32 (0)
         << uint32 ((uint32) - 1)
         << uint32 (need_face_now (pl, FINGER_DOWN))
         << data8  ("Apply this to see the items below")
         << uint16 (0)
         << uint8  (0)
         << uint32 (0);

      if (pl->ns->itemcmd == 2)
        sl << uint16 (0);

      dirty = true;
    }

  // now send out all items in the ring buffer in reverse order
  while (item_cnt)
    {
      --item_cnt;
      item_idx = (item_idx ? item_idx : FLOORBOX_PAGESIZE) - 1;
      object *item = items [item_idx];

      add_object_to_socklist (*pl->ns, sl, item->head_ ());

      dirty = true;

      // if packet got too large, send it and begin a new one
      if (sl.length () > MAXSOCKBUF - MAXITEMLEN)
        {
          pl->ns->send_packet (sl);

          sl.reset ();
          sl.printf ("item%d ", pl->ns->itemcmd);
          sl << uint32 (0);

          dirty = false;
        }
    }
  
  if (start_pos)
    {
      sl << uint32 (0x80000000 | (start_pos - FLOORBOX_PAGESIZE))
         << uint32 (0)
         << sint32 (-1)
         << uint32 (need_face_now (pl, FINGER_UP))
         << data8  ("Apply this to see the items higher up")
         << uint16 (0)
         << uint8  (0)
         << uint32 (0);

      if (pl->ns->itemcmd == 2)
        sl << uint16 (0);

      dirty = true;
    }

  if (dirty)
    pl->ns->send_packet (sl);
}

/**
 * Sends whole inventory.
 */
void
esrv_send_inventory (object *pl, object *op)
{
  if (!pl->contr->ns)//D
    return;

  int got_one = 0;

  pl->contr->ns->send_packet_printf ("delinv %d", op->count);

  packet sl;
  sl.printf ("item%d ", pl->contr->ns->itemcmd);

  sl << uint32 (op->count);

  for (object *tmp = op->inv; tmp; tmp = tmp->below)
    {
      object *head;

      if (tmp->head)
        head = tmp->head;
      else
        head = tmp;

      if (head->client_visible ())
        {
          add_object_to_socklist (*pl->contr->ns, sl, head);

          got_one++;

          /* IT is possible for players to accumulate a huge amount of
           * items (especially with some of the bags out there) to
           * overflow the buffer.  IF so, send multiple item commands.
           */
          if (sl.length () > MAXSOCKBUF - MAXITEMLEN)
            {
              pl->contr->ns->send_packet (sl);

              sl.reset ();
              sl.printf ("item%d ", pl->contr->ns->itemcmd);
              sl << uint32 (op->count);
              got_one = 0;
            }
        }
    }

  if (got_one)
    pl->contr->ns->send_packet (sl);
}

/**
 * Updates object *op for player *pl.
 *
 * flags is a list of values to update
 * to the client (as defined in newclient.h - might as well use the
 * same value both places.
 */
void
esrv_update_item (int flags, object *pl, object *op)
{
  /* If we have a request to send the player item, skip a few checks. */
  if (op != pl && !op->client_visible ())
    return;

  client *ns = pl->contr->ns;
  if (!ns)
    return;

  if (!QUERY_FLAG (op, FLAG_CLIENT_SENT))
    /* FLAG_CLIENT_SENT is debug only.  We are using it to see where
     * this is happening - we can set a breakpoint here in the debugger
     * and track back the call.
     */
    LOG (llevDebug | logBacktrace, "We have not sent item %s (%d)\n", &op->name, op->count);

  packet sl ("upditem");

  sl << uint8 (flags);

  op = op->head_ ();

  sl << uint32 (op->count);

  if (flags & UPD_LOCATION)
    sl << uint32 (op->env ? op->env->count : 0);

  if (flags & UPD_FLAGS)
    sl << uint32 (query_flags (op));

  if (flags & UPD_WEIGHT)
    {
      sint32 weight = op->flag [FLAG_NO_PICK] ? -1 : op->client_weight ();

      if (op)
        ns->last_weight = weight;

      sl << uint32 (weight);
    }

  if (flags & UPD_FACE)
    {
      ns->send_face (op->face, -50);
      ns->flush_fx ();
      sl << uint32 (op->face);
    }

  if (flags & UPD_NAME)
    {
      int len;
      const char *item_p;
      char item_n[MAX_BUF];

      if (!op->custom_name)
        {
          strncpy (item_n, query_base_name (op, 0), 127);
          item_n[127] = 0;
          len = strlen (item_n);
          item_p = query_base_name (op, 1);
        }
      else
        {
          strncpy (item_n, op->custom_name, 127);
          item_n[127] = 0;
          len = strlen (item_n);
          item_p = op->custom_name;
        }

      strncpy (item_n + len + 1, item_p, 127);
      item_n[254] = 0;
      len += strlen (item_n + 1 + len) + 1;

      sl << data8 (item_n, len);
    }

  if (flags & UPD_ANIM)
    sl << uint16 (op->animation_id);

  if (flags & UPD_ANIMSPEED)
    {
      int anim_speed = 0;

      if (QUERY_FLAG (op, FLAG_ANIMATE))
        {
          if (op->anim_speed)
            anim_speed = op->anim_speed;
          else
            {
              if (fabs (op->speed) < 0.001)
                anim_speed = 255;
              else if (fabs (op->speed) >= 1.0)
                anim_speed = 1;
              else
                anim_speed = (int) (1.0 / fabs (op->speed));
            }

          if (anim_speed > 255)
            anim_speed = 255;
        }

      sl << uint8  (anim_speed);
    }

  if (flags & UPD_NROF)
    sl << uint32 (op->nrof);

  pl->contr->ns->send_packet (sl);
}

/**
 * Sends item's info to player.
 */
void
esrv_send_item (object *pl, object *op)
{
  if (!pl->contr->ns)
    return;

  /* We only send 'visible' objects to the client, and sometimes the player */
  if (!op->client_visible () && op->type != PLAYER)
    return;

  packet sl;

  sl.printf ("item%d ", pl->contr->ns->itemcmd);

  op = op->head_ ();

  sl << uint32 (op->env ? op->env->count : 0);

  add_object_to_socklist (*pl->contr->ns, sl, op);

  pl->contr->ns->send_packet (sl);
  SET_FLAG (op, FLAG_CLIENT_SENT);
}

/**
 * Tells the client to delete an item.
 */
void
esrv_del_item (player *pl, int tag)
{
  if (!pl->ns)
    return;

  packet sl ("delitem");

  sl << uint32 (tag);

  pl->ns->send_packet (sl);
}


/*******************************************************************************
 *
 * Client has requested us to do something with an object.
 *
 ******************************************************************************/

/**
 * Takes a player and object count (tag) and returns the actual object
 * pointer, or null if it can't be found.
 */
object *
esrv_get_ob_from_count (object *pl, tag_t count)
{
  if (pl->count == count)
    return pl;

  for (object *op = pl->inv; op; op = op->below)
    if (op->count == count)
      return op;
    else if (op->type == CONTAINER && pl->container == op)
      for (object *tmp = op->inv; tmp; tmp = tmp->below)
        if (tmp->count == count)
          return tmp;

  for (object *op = GET_MAP_OB (pl->map, pl->x, pl->y); op; op = op->above)
    if (op->head && op->head->count == count)
      return op;
    else if (op->count == count)
      return op;
    else if (op->type == CONTAINER && pl->container == op)
      for (object *tmp = op->inv; tmp; tmp = tmp->below)
        if (tmp->count == count)
          return tmp;

#if 0
  /* If the high bit is set, player examined a pseudo object. */
  if (count & 0x80000000)
    return 0;
#endif

  return 0;
}

/** Client wants to examine some object.  So lets do so. */
void
ExamineCmd (char *buf, int len, player *pl)
{
  tag_t tag = atoi (buf);

  /* If the high bit is set, player applied a pseudo object. */
  if (tag & 0x80000000)
    {
      pl->ns->look_position = tag & 0x7fffffff;
      pl->ns->floorbox_update ();
      return;
    }

  object *op = esrv_get_ob_from_count (pl->ob, tag);

  if (!op)
    {
      LOG (llevDebug, "Player '%s' tried to examine the unknown object (%ld)\n", &pl->ob->name, tag);
      return;
    }

  examine (pl->ob, op);
}

/** Client wants to examine some object.  So lets do so. */
void
ExCmd (char *buf, int len, player *pl)
{
  tag_t tag = atoi (buf);

  if (object *op = esrv_get_ob_from_count (pl->ob, tag))
    {
      std::string s = op->describe (pl->ob);

      if (msg_is_special (s.c_str (), false))
        cfperl_expand_cfpod (pl, s);

      packet sl ("ex");
      sl << ber32 (tag) << s.c_str ();

      pl->ns->send_packet (sl);
    }
}

/** Client wants to apply some object. Lets do so. */
void
ApplyCmd (char *buf, int len, player *pl)
{
  tag_t tag = atoi (buf);

  /* If the high bit is set, player applied a pseudo object. */
  if (tag & 0x80000000)
    {
      pl->ns->look_position = tag & 0x7fffffff;
      pl->ns->floorbox_update ();
      return;
    }

  object *op = esrv_get_ob_from_count (pl->ob, tag);

  if (!op)
    {
      LOG (llevDebug, "Player '%s' tried to apply the unknown object (%d)\n", &pl->ob->name, tag);
      return;
    }

  player_apply (pl->ob, op, 0, 0);
}

/** Client wants to lock some object. Lets do so. */
void
LockItem (char *data, int len, player *pl)
{
  int flag = data[0];
  tag_t tag = net_uint32 ((uint8 *)data + 1);
  object *op = esrv_get_ob_from_count (pl->ob, tag);

  if (!op)
    {
      pl->failmsg ("Could not find object to lock/unlock");
      return;
    }

  if (!flag)
    CLEAR_FLAG (op, FLAG_INV_LOCKED);
  else
    SET_FLAG (op, FLAG_INV_LOCKED);

  esrv_update_item (UPD_FLAGS, pl->ob, op);
}

/** Client wants to mark some object.  Lets do so. */
void
MarkItem (char *data, int len, player *pl)
{
  tag_t tag = net_uint32 ((uint8 *)data);
  object *op = esrv_get_ob_from_count (pl->ob, tag);

  if (!op)
    {
      pl->failmsg ("Could not find object to mark");
      return;
    }

  pl->mark = op;
  pl->ob->statusmsg (format ("Marked item %s", query_name (op)));
}

/**
 * look_at prints items on the specified square.
 *
 * [ removed EARTHWALL check and added check for containers inventory.
 *   Tero.Haatanen@lut.fi ]
 */
static void
look_at (player *pl, int dx, int dy)
{
  dynbuf_text buf;
  object *ob = pl->ob;

  if (!pl->observe->map)
    return;

  mapxy pos (pl->observe);
  pos.move (dx, dy);

  if (pos.normalise ())
    for (object *tmp = pos->top; tmp; tmp = tmp->below)
      {
        if (tmp->invisible && !QUERY_FLAG (ob, FLAG_WIZ))
          continue;

        if (QUERY_FLAG (ob, FLAG_WIZ))
          buf.printf ("- %s (%d).\n", query_name (tmp), tmp->count);
        else
          buf.printf ("- %s.\n", query_name (tmp));

        object *head = tmp->head_ ();

        if (head->inv)
          if ((head->type != CONTAINER && head->type != FLESH)
              || QUERY_FLAG (ob, FLAG_WIZ))
            buf << head->query_inventory (ob, "  ");

        if (QUERY_FLAG (tmp, FLAG_IS_FLOOR) && !QUERY_FLAG (ob, FLAG_WIZ))        /* don't continue under the floor */
          break;
      }

  if (buf.empty ())
    pl->failmsg ("You see nothing there.");
  else
    pl->infobox (MSG_CHANNEL ("lookat"), buf);
}

/** Client wants to look at some object.  Lets do so. */
void
LookAt (char *buf, int len, player *pl)
{
  char *cp;

  int dx = atoi (buf);
  if (!(cp = strchr (buf, ' ')))
    return;

  int dy = atoi (cp);

  if (player *opl = pl->observe->contr)
    if (client *ns = opl->ns)
      {
        if (fabs (dx) > ns->mapx / 2 || fabs (dy) > ns->mapy / 2)
          return;

        if (opl->blocked_los[dx + ns->mapx / 2][dy + ns->mapy / 2])
          return;
      }

  look_at (pl, dx, dy);
}

/** Move an object to a new location */
void
esrv_move_object (object *pl, tag_t to, tag_t tag, long nrof)
{
  object *op = esrv_get_ob_from_count (pl, tag);
  if (!op)
    {
      LOG (llevDebug, "Player '%s' tried to move an unknown object (%ld)\n", &pl->name, tag);
      return;
    }

  if (!to)
    {                           /* drop it to the ground */
      if (op->map && !op->env)
        return;

      /* If it is an active container, then we should drop all objects
       * in the container and not the container itself.
       */
      if (op->inv && QUERY_FLAG (op, FLAG_APPLIED))
        {
          int cnt = MAX_ITEM_PER_DROP;

          for (object *current = op->inv; current && cnt--; )
            {
              object *next = current->below;
              drop_object (pl, current, 0);
              current = next;
            }

          if (cnt <= 0)
            op->failmsg ("Only dropped some items, can't drop that many items at once.");
        }
      else
        drop_object (pl, op, nrof);

      return;
    }
  else if (to == pl->count)
    {                           /* pick it up to the inventory */
      /* return if player has already picked it up */
      if (op->env == pl)
        return;

      pl->contr->count = nrof;
      pick_up (pl, op);
      return;
    }

  object *env = esrv_get_ob_from_count (pl, to);
  if (!env)
    {
      LOG (llevDebug, "Player '%s' tried to move object to the unknown location (%d)\n", &pl->name, to);
      return;
    }

  /* put_object_in_sack presumes that necessary sanity checking
   * has already been done (eg, it can be picked up and fits in
   * in a sack, so check for those things.  We should also check
   * an make sure env is in fact a container for that matter.
   */
  if (env->type == CONTAINER
      && can_pick (pl, op)
      && sack_can_hold (pl, env, op, nrof))
    put_object_in_sack (pl, env, op, nrof);
}

